﻿/*========================================================================
        Copyright (C) 2006 by Geir-Tore Lindsve, Torbjørn Meistad 
        and Yngve Raudberget, hereby refered to as "the authors".
                        All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for NONCOMMERCIAL RESEARCH AND EDUCATIONAL PURPOSES, provided 
    that (1) the above copyright notice and the following two paragraphs 
    appear in all copies of the source code and (2) redistributions, 
    including without limitation binaries, reproduce these notices in 
    the supporting documentation. 

    IN NO EVENT SHALL THE AUTHORS, OR DISTRIBUTORS OF THIS SOFTWARE 
    BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, 
    OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE 
    AND ITS DOCUMENTATION, EVEN IF THE AUTHORS OR ANY OF THE ABOVE 
    PARTIES HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    THE AUTHORS SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
    BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
    FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
    ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE NO
    OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
    MODIFICATIONS.
========================================================================*/

using System;
using System.Collections.Generic;
using System.Text;
using CLab.Exceptions;

namespace CLab.CSP
{
    /// <summary>
    /// A class for representing boolean types in the CSP part of CLab.
    /// The different possible domain values for a range are encoded as 
    /// a range with an end point. The start point is assumed always to be 0. 
    /// Implements the abstract CSPType class.
    /// </summary>
    public class CSPTypeBool : CSPType
    {

        /// <summary>
        /// Initializes a new instance of the <see cref="T:CSPTypeBool"/> class.
        /// </summary>
        /// <param name="typeName">Name of the type.</param>
        public CSPTypeBool(String typeName) 
            : base(typeName, 1)
        {
        }

        /// <summary>
        /// Returns the <see cref="F:CPTypes.bool_type"></see> constant.
        /// </summary>
        /// <returns>The <see cref="F:CPTypes.bool_type"></see> constant.</returns>
        public override CPTypes TypeFormat()
        {
            return CPTypes.bool_type;
        }

        /// <summary>
        /// Gets the string representation of an int within the range of a type.
        /// </summary>
        /// <param name="domainInt">An integer representing a domain value.</param>
        /// <returns>
        /// The string representation of the domain value.
        /// </returns>
        /// <exception cref="CLab.Exceptions.ClabException"/>
        public override String GetStringRepOfDomainInt(int domainInt)
        {
            if (domainInt < 2 && domainInt > -1)
            {
                return "" + domainInt;
            }
            throw new ClabException("Invalid domain int");
        }

        /// <summary>
        /// Gets the int representation of the string representation of a domain value.
        /// </summary>
        /// <param name="domainString">The string representation of a domain value.</param>
        /// <returns>
        /// The int representation of the domain value.
        /// </returns>
        /// <exception cref="T:CLab.Exceptions.ClabException">Exception thrown when string representation is invalid</exception>
        public override int GetIntRepOfDomainString(String domainString)
        {
            try
            {
                int domainInt = Int32.Parse(domainString);
                if (domainInt == 0 || domainInt == 1) return domainInt;
                else return 1;
            }
            catch (Exception e)
            {
                throw new ClabException(domainString + " is not a valid string representation of a domain value for a csp boolean type", e);
            }
        }

        /// <summary>
        /// Returns a <see cref="T:System.String"></see> that represents the current <see cref="T:CLab.CSP.CSPTypeBool"></see>.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String"></see> that represents the current <see cref="T:CLab.CSP.CSPTypeBool"></see>.
        /// </returns>
        public override string ToString()
        {
            return "CSPTypeBool - name: " +base.ToString();
        }

     

    }
}
