﻿/*========================================================================
        Copyright (C) 2006 by Geir-Tore Lindsve, Torbjørn Meistad 
        and Yngve Raudberget, hereby refered to as "the authors".
                        All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for NONCOMMERCIAL RESEARCH AND EDUCATIONAL PURPOSES, provided 
    that (1) the above copyright notice and the following two paragraphs 
    appear in all copies of the source code and (2) redistributions, 
    including without limitation binaries, reproduce these notices in 
    the supporting documentation. 

    IN NO EVENT SHALL THE AUTHORS, OR DISTRIBUTORS OF THIS SOFTWARE 
    BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, 
    OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE 
    AND ITS DOCUMENTATION, EVEN IF THE AUTHORS OR ANY OF THE ABOVE 
    PARTIES HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    THE AUTHORS SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
    BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
    FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
    ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE NO
    OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
    MODIFICATIONS.
========================================================================*/

using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;

using CLab.Exceptions;

namespace CLab.CSP
{
    /// <summary>
    /// A class for representing enumeration types in the CSP part of CLab.
    /// The different possible domain values for an enumeration are encoded as 
    /// a range with an end point. The start point is assumed always to be 0.
    /// Implements the abstract CSPType class.
    /// </summary>
    public class CSPTypeEnum : CSPType
    {
        private Hashtable constantToID;
        private List<String> idToConstant;
        private int lastID = -1;
        private StringComparer stringComparer;

        /// <summary>
        /// Initializes a new instance of the <see cref="T:CSPTypeEnumerator"/> class.
        /// </summary>
        /// <param name="typeName">Name of the type.</param>
        public CSPTypeEnum(String typeName) 
            : base(typeName, -1)
        {
            stringComparer = StringComparer.Create(System.Globalization.CultureInfo.CurrentCulture, true);
            constantToID = new Hashtable(stringComparer);
            idToConstant = new List<String>();
        }

        /// <summary>
        /// Adds a new enumeration domain value (constant). 
        /// This method makes a mapping from string representation 
        /// to int representation and vice versa.
        /// </summary>
        /// <param name="constant">The constant.</param>
        /// <returns></returns>
        public int AddConstant(String constant)
        {
            lastID++;
            constantToID[constant] = lastID;
            idToConstant.Add(constant);
            base.End = lastID;
            return lastID;
        }

        /// <summary>
        /// Returns the <see cref="F:CPTypes.enum_type"></see> constant.
        /// </summary>
        /// <returns>The <see cref="F:CPTypes.enum_type"></see> constant.</returns>
        public override CPTypes TypeFormat()
        {
            return CPTypes.enum_type;
        }

        /// <summary>
        /// Gets the string representation of an int within the range of a type.
        /// Uses the internal mapping from int to string.
        /// </summary>
        /// <param name="domainInt">An integer representing an enumeration domain value (constant).</param>
        /// <returns>
        /// The string representation of the domain value.
        /// </returns>
        /// <exception cref="CLab.Exceptions.ClabException"/>
        public override String GetStringRepOfDomainInt(int domainInt)
        {
            if (domainInt > -1 && domainInt < idToConstant.Count)
            {
                return idToConstant[domainInt];
            }
            throw new ClabException("Invalid domain int");
        }

        /// <summary>
        /// Gets the int representation of the string representation of a domain value.
        /// Uses the internal mapping from string to int.
        /// </summary>
        /// <param name="domainString">The string representation of a domain value.</param>
        /// <returns>
        /// The int representation of the domain value.
        /// </returns>
        /// <exception cref="CLab.Exceptions.ClabException"/>
        public override int GetIntRepOfDomainString(String domainString)
        {
            try
            {
                return (int)constantToID[domainString];
            }
            catch (Exception e)
            {
                throw new ClabException("Invalid domain string", e);
            }
        }


        /// <summary>
        /// Returns a <see cref="T:System.String"></see> that represents the current <see cref="T:CLab.CSP.CSPTypeEnum"></see>.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String"></see> that represents the current <see cref="T:CLab.CSP.CSPTypeEnum"></see>.
        /// </returns>
        public override string ToString()
        {
            String retur = "CSPTypeEnumerator - name: " +base.ToString() + "\n";
            IDictionaryEnumerator enumer = constantToID.GetEnumerator();
            while(enumer.MoveNext()){
                retur += enumer.Key + ": " + enumer.Value + " - ";
            }
            return retur;
        }
    }
}
