﻿/*========================================================================
        Copyright (C) 2006 by Geir-Tore Lindsve, Torbjørn Meistad 
        and Yngve Raudberget, hereby refered to as "the authors".
                        All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for NONCOMMERCIAL RESEARCH AND EDUCATIONAL PURPOSES, provided 
    that (1) the above copyright notice and the following two paragraphs 
    appear in all copies of the source code and (2) redistributions, 
    including without limitation binaries, reproduce these notices in 
    the supporting documentation. 

    IN NO EVENT SHALL THE AUTHORS, OR DISTRIBUTORS OF THIS SOFTWARE 
    BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, 
    OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE 
    AND ITS DOCUMENTATION, EVEN IF THE AUTHORS OR ANY OF THE ABOVE 
    PARTIES HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    THE AUTHORS SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
    BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
    FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
    ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE NO
    OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
    MODIFICATIONS.
========================================================================*/

using System;
using System.Collections.Generic;
using System.Text;
using CLab.Exceptions;

namespace CLab.CSP
{
    /// <summary>
    /// A class for representing range types in the CSP part of CLab.
    /// The different possible domain values for a range are encoded as 
    /// a range with a start point and an end point. Implements the abstract
    /// CSPType class.
    /// </summary>
    public class CSPTypeRange : CSPType
    {
        private int start;
       
        public CSPTypeRange(String typeName, int start, int end) 
            : base(typeName, end)
        {
            this.start = start;
        }

        /// <summary>
        /// Gets the start point of the range.
        /// </summary>
        /// <value>The start point.</value>
        public int Start
        {
            get { return start; }
        }

        /// <summary>
        /// Returns the <see cref="F:CPTypes.range_type"></see> constant..
        /// </summary>
        /// <returns>The <see cref="F:CPTypes.range_type"></see> constant.</returns>
        public override CPTypes TypeFormat()
        {
            return CPTypes.range_type;
        }

        /// <summary>
        /// Gets the string representation of an int within the range.
        /// </summary>
        /// <param name="domainInt">The domain int.</param>
        /// <returns>The string representation of the domain int.</returns>
        /// <exception cref="CLab.Exceptions.ClabException"/>
        public override String GetStringRepOfDomainInt(int domainInt)
        {
            if (domainInt >= start && domainInt <= End)
            {
                return "" + domainInt;
            }
            throw new ClabException("Invalid domain int");
        }

        /// <summary>
        /// Gets the int representation of the string representation of a domain value.
        /// </summary>
        /// <param name="domainString">The string representation of a domain value.</param>
        /// <returns>
        /// The int representation of the domain value.
        /// </returns>
        /// <exception cref="CLab.Exceptions.ClabException">Exception thrown when string representation is invalid</exception>
        public override int GetIntRepOfDomainString(String domainString)
        {
             try
            {
                int domainInt = Int32.Parse(domainString);
                if (domainInt >= start & domainInt <= End) return domainInt;
                else throw new Exception("" + domainInt + " is not valid");
             }
            catch (Exception e)
            {
                throw new ClabException(domainString +" is not a valid string representation of a domain value for a csp range type", e);
            }
        }

        /// <summary>
        /// Returns a <see cref="T:System.String"></see> that represents the current <see cref="T:CLab.CSP.CSPTypeRange"></see>.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String"></see> that represents the current <see cref="T:CLab.CSP.CSPTypeRange"></see>.
        /// </returns>
        public override string ToString()
        {
            return "CSPTypeRange - name: " + base.ToString() + " start: " + start + " end: " + base.End;
        }
    }
}