/*========================================================================
        Copyright (C) 2006 by Geir-Tore Lindsve, Torbjrn Meistad 
        and Yngve Raudberget, hereby refered to as "the authors".
                        All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for NONCOMMERCIAL RESEARCH AND EDUCATIONAL PURPOSES, provided 
    that (1) the above copyright notice and the following two paragraphs 
    appear in all copies of the source code and (2) redistributions, 
    including without limitation binaries, reproduce these notices in 
    the supporting documentation. 

    IN NO EVENT SHALL THE AUTHORS, OR DISTRIBUTORS OF THIS SOFTWARE 
    BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, 
    OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE 
    AND ITS DOCUMENTATION, EVEN IF THE AUTHORS OR ANY OF THE ABOVE 
    PARTIES HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    THE AUTHORS SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
    BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
    FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
    ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE NO
    OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
    MODIFICATIONS.
========================================================================*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using CLab.Exceptions;
using CLab.Parsers;

namespace CLab.Data
{
    /// <summary>
    /// The CP object contains the internal data representation, parsed from the input XML file
    /// </summary>
    public class CP
    {
        private List<Type> types = new List<Type>();
        private List<Variable> variables = new List<Variable>();
        private List<Expression> rules = new List<Expression>();
        private String description = "";
        private String author = "";
        private String date = "";
        private ClabXmlParser parser;

        /// <summary>
        /// Initializes a new instance of the <see cref="T:CP"/> class.
        /// </summary>
        /// <param name="xmlFilename">The XML filename.</param>
        /// <exception cref="CLab.Exceptions.ClabException"></exception>
        public CP(String xmlFilename)
        {
            parser = new ClabXmlParser(this);
            try
            {
                parser.Parse(xmlFilename);
            }
            catch (ClabXMLParserException e)
            {
                throw new ClabException(e.Message, e);
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="T:CP"/> class.
        /// </summary>
        /// <param name="stream">The stream.</param>
        public CP(Stream stream)
        {
            parser = new ClabXmlParser(this);
            try
            {
                parser.Parse(stream);
            }
            catch (ClabXMLParserException e)
            {
                throw new ClabException(e.Message, e);
            }
        }

        /// <summary>
        /// Optional header information in the XML file.
        /// Description of the XML data (optional metadata)
        /// </summary>
        public String Description
        {
            get { return description; }
            set { description = value; }
        }

        /// <summary>
        /// Optional header information in the XML file.
        /// Author of the XML data (optional metadata)
        /// </summary>
        public String Author
        {
            get {return author; }
            set {author = value; }
        }

        /// <summary>
        /// Optional header information in the XML file.
        /// Timestamp of the XML data (optional metadata)
        /// </summary>
        /// <value>The date.</value>
        public String Date 
        {
            get { return date; }
            set { date = value; }
        }

        /// <summary>
        /// Gets the list of type objects in the internal data representation
        /// </summary>
        /// <value>The types.</value>
        public List<Type> Types
        {
            get { return types; }
        }

        /// <summary>
        /// Gets the List of variable objects in the internal data representation.
        /// </summary>
        /// <value>The list of variables.</value>
        public List<Variable> Variables
        {
            get { return variables; }
        }

        /// <summary>
        /// Gets the list of rule (expression) objects in the internal data representation
        /// </summary>
        /// <value>The rules.</value>
        public List<Expression> Rules
        {
            get { return rules; }
        }

        /// <summary>
        /// Adds a type object to the internal data representation
        /// </summary>
        /// <param name="newType">The type object to add</param>
        public void AddType(Type newType)
        {
            types.Add(newType);
        }

        /// <summary>
        /// Adds a variable object to the internal data representation
        /// </summary>
        /// <param name="newVariable">The variable object to add</param>
        public void AddVariable(Variable newVariable)
        {
            variables.Add(newVariable);
        }

        /// <summary>
        /// Adds a rule (expression) object to the internal data representation
        /// </summary>
        /// <param name="newRule">The rule to add</param>
        public void AddRule(Expression newRule)
        {
            rules.Add(newRule);
        }
    }
}
