/*========================================================================
        Copyright (C) 2006 by Geir-Tore Lindsve, Torbjrn Meistad 
        and Yngve Raudberget, hereby refered to as "the authors".
                        All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for NONCOMMERCIAL RESEARCH AND EDUCATIONAL PURPOSES, provided 
    that (1) the above copyright notice and the following two paragraphs 
    appear in all copies of the source code and (2) redistributions, 
    including without limitation binaries, reproduce these notices in 
    the supporting documentation. 

    IN NO EVENT SHALL THE AUTHORS, OR DISTRIBUTORS OF THIS SOFTWARE 
    BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, 
    OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE 
    AND ITS DOCUMENTATION, EVEN IF THE AUTHORS OR ANY OF THE ABOVE 
    PARTIES HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    THE AUTHORS SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
    BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
    FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
    ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE NO
    OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
    MODIFICATIONS.
========================================================================*/

using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using Iesi.Collections;

namespace Casper.Algorithm
{
    /// <summary>
    /// The class which stores the information 
    /// of a variable and its domaine values. Used by 
    /// <see cref="T:GeneralizedLookahead"></see>.
    /// </summary>
    public class LookAheadVarDom
    {
        private int varID;
        private Set domainValues;
        private Set removedValues;
        private IEnumerator domainEnumerator;
        private int? selectedValue;

        /// <summary>
        /// Initializes a new instance of the <see cref="T:LookAheadVarDom"/> class.
        /// </summary>
        /// <param name="varID">The variable ID.</param>
        /// <param name="domainValues">The domain values.</param>
        public LookAheadVarDom(int varID, Set domainValues)
        {
            removedValues = new HybridSet();
            this.varID = varID;
            this.domainValues = domainValues;
            domainEnumerator = domainValues.GetEnumerator();
        }

        /// <summary>
        /// Gets og sets the variable ID.
        /// </summary>
        /// <value>The variable ID.</value>
        public int VarID
        {
            get { return varID; }
            set { varID = value; }
        }

        /// <summary>
        /// Gets or sets the domain values.
        /// </summary>
        /// <value>The domain values.</value>
        public Set DomainValues
        {
            get { return domainValues; }
            set
            {
                domainValues = value;
                domainEnumerator = domainValues.GetEnumerator();
            }
        }

        /// <summary>
        /// Gets the current selected value of the variable
        /// </summary>
        /// <value></value>
        public int? SelectedValue
        {
            get { return selectedValue; }
        }

        /// <summary>
        /// Gets the next value. The domain value ordering is ascending.
        /// Jumps over values which is added to the removed set by the
        /// <see cref="AddToRemoved"></see> method.
        /// </summary>
        /// <param name="startFromBeginning">If set to <c>true</c>, start from the beginning.</param>
        /// <returns></returns>
        public int? GetNextValue(Boolean startFromBeginning)
        {
            if (startFromBeginning)
            {
                domainEnumerator.Reset();
            }
            while (domainEnumerator.MoveNext())
            {
                int nextVal = (int)domainEnumerator.Current;
                selectedValue = nextVal;
                if (removedValues.Contains(selectedValue))
                {
                    continue;
                }
                return nextVal;
            }
            selectedValue = null;
            return null;
        }

        /// <summary>
        /// Adds a value to a set of removed values.
        /// The value is not removed from the domain values before
        /// <see cref="UpdateDomainWithRemoved"></see> is run.
        /// </summary>
        /// <param name="val">The value.</param>
        public void AddToRemoved(int val)
        {
            removedValues.Add(val);
        }

        /// <summary>
        /// Updates the domain with removed values in the removed set.
        /// The removed set is emptied.
        /// </summary>
        public void UpdateDomainWithRemoved()
        {
            domainValues = (Set)domainValues.Minus(removedValues);
            domainEnumerator = domainValues.GetEnumerator();
            removedValues.Clear();
        }

        /// <summary>
        /// Method used by the algorithm to reset the domain during backtracking.
        /// </summary>
        /// <param name="removedVals">The removed values which should be reset.</param>
        public void ResetDomain(Set removedVals)
        {
            domainValues = (Set) domainValues.Union(removedVals);
            domainEnumerator = domainValues.GetEnumerator();
            removedVals.Clear();
        }

        /// <summary>
        /// Gets the domain size.
        /// </summary>
        /// <returns>The domain size.</returns>
        public int DomainSize()
        {
            return domainValues.Count;
        }
    }
}
